/*
 * assem.c
 *
 * Ein konfigurierbarer Assembler.
 *
 * Durch einfache Aenderungen in scan.lex, tokens.h und command.h
 * kann man leicht eine eigene Maschinensprache definieren.
 * Um die wirkliche Zahlenrepresentation zu erzeugen muss man
 * dann noch die zwei Funktionen ReadData() und DoCommand()
 * (siehe am Ende dieses Files) implementieren.
 *
 * v1.0  5. Mai 95  Georg He"smann.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <varargs.h>
#include <string.h>

#include <assert.h>

int printf();
void rewind();
int fprintf();
int strtol();
int vfprintf();

#include "tokens.h"
#include "command.h"
#include "symtab.h"

extern int    yyLineNumber;
extern char * yyFileName;
extern char * yytext;
extern char * xxStrBuf;
extern int    yyeof;

FILE * yyin, * yyout;

extern void yyrestart(FILE *);
extern int  yylex(void);
extern void unputToken(void);

int Pass1(void);
int Pass2(void);
int ParseArgs(int comNumber);
int ReadData(void);
int ParseEOArg(int * numArg, int * endCom);
int ParseLBR(int comNumber, int * numArg, int * endCom, struct ArgAdrMode args[]);
int ParseMINUS(int comNumber, int * numArg, int * endCom, struct ArgAdrMode args[]);
int ParseHEX(int comNumber, int * numArg, int * endCom, struct ArgAdrMode args[]);
int DoCommand(int comNumber, struct ArgAdrMode args[]);
void ParseError();	/* K&R Prototyp */
int isValidAdrMode(struct ValidAdrModes * valid, enum AdrMode mode);


int ProgStartAddress = 0;	/* wohin soll das Programm assembliert werden */
int CurrentAddress;		/* wo wird der aktuelle Befehl abgelegt */

int main(int argc, char * argv[])
{
  int error = 0;
  
  if (argc == 2) {
    yyin = fopen(argv[1], "r");
    yyFileName = argv[1];
  }
  else {
    printf("usage: %s asmfile\n", argv[0]);
    exit(1);
  }

  assert(yyin);

  yyout = stderr;

  error = Pass1();	/* aufsammeln aller Labels und deren Position bestimmen */
  if (!error) {
    yyLineNumber = 1;
    yyeof = 0;
    rewind(yyin);
    yyrestart(yyin);
    Pass2();
  }
  
  return error;
}


int Pass1(void)
{
  int error = 0;
  int nextProgAddress = 0;
  int xxresult;
  int inDataBlk = 0;		/* wird gerade eine DATA Zeile gelesen? */
  
  do {
    xxresult = yylex();
    if (xxresult != xx_EOF) {

      if (inDataBlk) {
	if (xxresult == xx_semi) inDataBlk = 0;
	else {
	  switch (xxresult) {
	  case xx_com:
	    break;
	  case xx_NUM:
	  case xx_HEX:
	    nextProgAddress++;					/* immer ein 16-Bit Wort */
	    break;
	  case xx_STRING:
	    nextProgAddress += (strlen(xxStrBuf) + 1) / 2;	/* auf 16 Bit aufrunden */
	    break;
	  default:
	    ParseError("Unerwartetes Token (%d) im Data Block.", xxresult);
	    error = 1;
	    inDataBlk = 0;
	  }
	}
      }
      else if (xxresult < MAXCOMMAND) {
      	/* Es ist ein Befehl, kein sonstiges Token */
	nextProgAddress += ComTab[xxresult].co_BefLength;
      }
      else {
	/* einfaches Zeichen-Token oder Nichtterminal */
	if (xxresult == xx_STARTADR) {
	  xxresult = yylex();
	  
	  switch (xxresult) {
	  case xx_NUM:
	    ProgStartAddress = strtol(yytext, NULL, 10);
	    break;
	  case xx_HEX:
	    ProgStartAddress = strtol(yytext+1, NULL, 16);
	    break;
	  default:
	    ParseError("Ungueltige StartAdresse.");
	    error = 1;
	  }

	  if (!error) {
	    xxresult = yylex();
	    if (xxresult != xx_semi) {
	      ParseError("Semikolon nach .STARTADR fehlt");
	      error = 1;
	    }
	  }
	}
	else if (xxresult == xx_DATA) {
	  inDataBlk = 1;
	}
	else if (xxresult == xx_LABEL) {
	  int temp;
	  yytext[strlen(yytext)-1] = '\0';	/* weg mit dem ':' */
	  if (!strcmp(yytext, "PC")) {
	    ParseError("\"PC\" ist ein reserviertes Wort.");
	    error = 1;
	  }
	  else {
	    temp = InsertEntry(yytext, nextProgAddress);
	    if (temp) {
	      ParseError("Label %s war bereits definiert.", yytext);
	      error = 1;
	    }
	  }
	}
      }
    }
  } while (xxresult != xx_EOF);

  if (inDataBlk) {
    ParseError("Nach .DATA fehlt ein Semikolon.");
    error = 1;
  }
  
  return error;
}


int Pass2(void)
{
  int xxresult;
  int error = 0;

  CurrentAddress = ProgStartAddress;

  do {					/* solange nicht EOF */
    xxresult = yylex();			/* lese naechstes Token */

    /* wir sind am Anfang einer Zeile, hier sind nur folgende Tokens erlaubt: */
    if (xxresult > xxDONE && xxresult < MAXCOMMAND) {
      /* dann ist das Token ein Befehl */
      error = ParseArgs(xxresult);
      
      /* Befehl ist komplett eingelesen und assembliert worden. */
      /* Nun Folgeadresse bestimmen. */
      if (!error) {
	CurrentAddress += ComTab[xxresult].co_BefLength;
      }
    }
    else {
      switch (xxresult) {
      case xx_EOF:
	break;
      case xx_LABEL:
	/* Label Definition. Wird in Pass2 nichtmehr benoetigt. */
	break;
      case xx_DATA:
	/* Daten Definition */
	error = ReadData();
	break;
      case xx_STARTADR:
	/* Startadresse wurde bereits in Pass 1 gelesen */
	xxresult = yylex();
	assert(xxresult == xx_NUM || xxresult == xx_HEX);
	xxresult = yylex();
	assert(xxresult == xx_semi);
	break;
      default:
	ParseError("Ungueltiges Token (%d) am Zeilenbeginn.", xxresult);
	break;
      }
    }
  }while (xxresult != xx_EOF && !error);

  return error;
}

int ParseArgs(int comNumber)
{
  int error = 0;
  int xxresult;
  int endCom = 0;	/* wurden alle Argumente eingelesen? (";" gelesen) */
  int tmp;
  int  numArg = 0;
  struct ArgAdrMode args[MAXARGS];

  while (!endCom && !error) {
    xxresult = yylex();			/* das naechste Token einlesen */
  
    switch (xxresult) {
    case xx_ID:			/* ID */
      if (!ExistEntry(yytext)) {
	/* kein Label, also ein Register */
	args[numArg].aam_mode = am_direct;
	args[numArg].aam_ID = strdup(yytext);
      }
      else {
	/* ein Label, also eine absolute Adresse */
	args[numArg].aam_mode = am_absolut;
	(void)GetEntryData(yytext, &tmp);
	args[numArg].aam_hex = tmp + ProgStartAddress;
      }
      tmp = isValidAdrMode(&ComTab[comNumber].co_AdrModesArg[numArg], args[numArg].aam_mode);
      assert(tmp != -1);
      if (!tmp) {
	ParseError("Ungueltiger Adressmodus fuer Argument %d.", numArg+1);
	error = 1;
      }
      else {
	/* nach einem Argument folgt entweder ein "," oder ein ";" */
	error = ParseEOArg(&numArg, &endCom);
      }
      break;

    case xx_lbr:		/* (ID) oder (ID)+ */
      error = ParseLBR(comNumber, &numArg, &endCom, args);
      break;
      
    case xx_minus:		/* -(ID) */
      error = ParseMINUS(comNumber, &numArg, &endCom, args);
      break;
      
    case xx_HEX:		/* $xx(ID) oder $xx(PC) (PC-relativ) */
      args[numArg].aam_mode = am_riadrdist;
      args[numArg].aam_hex = strtol(yytext+1, NULL, 16);
      tmp = isValidAdrMode(&ComTab[comNumber].co_AdrModesArg[numArg], args[numArg].aam_mode);
      if (!tmp) {
	ParseError("Ungueltiger Adressmodus fuer Argument %d.", numArg+1);
	error = 1;
      }
      else {
	/* noch das "(ID)" lesen */
	error = ParseHEX(comNumber, &numArg, &endCom, args);
      }
      break;
      
    case xx_ConstHEX:		/* #$xx */
      args[numArg].aam_mode = am_immed;
      args[numArg].aam_hex = strtol(yytext+2, NULL, 16);
      tmp = isValidAdrMode(&ComTab[comNumber].co_AdrModesArg[numArg], args[numArg].aam_mode);
      assert(tmp != -1);
      if (!tmp) {
	ParseError("Ungueltiger Adressmodus fuer Argument %d.", numArg+1);
	error = 1;
      }
      else {
	/* nach einem Argument folgt entweder ein "," oder ein ";" */
	error = ParseEOArg(&numArg, &endCom);
      }
      break;

    case xx_semi:		/* no args */
      endCom = 1;
      break;
      
    default:
      ParseError("Unerwartetes Token (%d) in der Argumentliste.", xxresult);
      error = 1;
      break;
    }
  }

  if (endCom) {
    /* Befehlszeile wurde komplett eingelesen */
    if (numArg == ComTab[comNumber].co_NumArgs) {
      DoCommand(comNumber, args);
    }
    else {
      ParseError("%d anstatt von %d Argumenten.", numArg, ComTab[comNumber].co_NumArgs);
      error = 1;
    }
  }
  
  return error;
}


int ParseLBR(int comNumber, int * numArg, int * endCom, struct ArgAdrMode args[])
{
  int error = 0;
  int xxresult;
  int tmp;

  xxresult = yylex();
  if (xxresult != xx_ID) {
    ParseError("Register/Label erwartet.");
    error = 1;
  }
  else {
    /* es wurde ein ID gelesen */
    if (ExistEntry(yytext)) {
      /* huch, es ist ein Label? */
      ParseError("Label anstatt eines Registers gelesen.");
      error = 1;
    }
    else {
      args[*numArg].aam_ID = strdup(yytext);
      xxresult = yylex();
      if (xxresult != xx_rbr) {
	ParseError("Klammer zu erwartet.");
	error = 1;
      }
      else {
	xxresult = yylex();
	if (xxresult == xx_add) {
	  args[*numArg].aam_mode = am_ripostinc;
	  tmp = isValidAdrMode(&ComTab[comNumber].co_AdrModesArg[*numArg], args[*numArg].aam_mode);
	  assert(tmp != -1);
	  if (!tmp) {
	    ParseError("Ungueltiger Adressmodus fuer Argument %d.", *numArg + 1);
	    error = 1;
	  }
	}
	else {
	  args[*numArg].aam_mode = am_regind;
	  tmp = isValidAdrMode(&ComTab[comNumber].co_AdrModesArg[*numArg], args[*numArg].aam_mode);
	  assert(tmp != -1);
	  if (!tmp) {
	    ParseError("Ungueltiger Adressmodus fuer Argument %d.", *numArg);
	    error = 1;
	  }
	  
	  /* letztes Zeichen wieder zurueck */
	  unputToken();
	}
	error = ParseEOArg(numArg, endCom);
      }
    }
  }
  
  return error;
}

int ParseMINUS(int comNumber, int * numArg, int * endCom, struct ArgAdrMode args[])
{
  int error = 0;
  int xxresult;

  xxresult = yylex();
  if (xxresult != xx_lbr) {
    ParseError("Klammer auf erwartet.");
    error = 1;
  }
  else {
    xxresult = yylex();
    if (xxresult != xx_ID) {
      ParseError("Register erwartet.");
      error = 1;
    }
    else {
      if (ExistEntry(yytext)) {
	ParseError("Label anstatt eines Registers gelesen.");
	error = 1;
      }
      else {
	args[*numArg].aam_ID = strdup(yytext);
	xxresult = yylex();
	if (xxresult != xx_rbr) {
	  ParseError("Klammer zu erwartet.");
	  error = 1;
	}
	else {
	  error = ParseEOArg(numArg, endCom);
	}
      }
    }
  }
  return error;
}

int ParseHEX(int comNumber, int * numArg, int * endCom, struct ArgAdrMode args[])
{
  int error = 0;
  int xxresult;

  xxresult = yylex();
  if (xxresult != xx_lbr) {
    ParseError("Klammer auf erwartet.");
    error = 1;
  }
  else {
    xxresult = yylex();
    if (xxresult != xx_ID) {
      ParseError("Register erwartet.");
      error = 1;
    }
    else {
      if (ExistEntry(yytext)) {
	ParseError("Label anstatt eines Registers gelesen.");
	error = 1;
      }
      else {
	if (!strcmp(yytext, "PC")) {
	  args[*numArg].aam_mode = am_pcrel;
	}
	else {
	  args[*numArg].aam_mode = am_riadrdist;
	}
	args[*numArg].aam_ID = strdup(yytext);
	xxresult = yylex();
	if (xxresult != xx_rbr) {
	  ParseError("Klammer zu erwartet.");
	  error = 1;
	}
	else {
	  error = ParseEOArg(numArg, endCom);
	}
      }
    }
  }

  return error;
}

int ParseEOArg(int * numArg, int * endCom)
{
  int error = 0;
  int xxresult;
  
  xxresult = yylex();
  
  switch (xxresult) {
  case xx_com:
    (*numArg)++;	/* naechstes Arg wird gelesen */
    break;
  case xx_semi:
    (*numArg)++;
    *endCom = 1;	/* Befehl ist komplett eingelesen */
    break;
  case xx_EOF:
    ParseError("Fileende mitten in einem Kommando.");
    error = 1;
    break;
  default:
    ParseError("Unerwartetes Token (%d).", xxresult);
    error = 1;
    break;
  }

  return error;
}



void ParseError(va_alist)
     va_dcl
{
  va_list args;
  char * fmt;
  
  va_start(args);
  fprintf(stderr, "%s:%d: ", yyFileName, yyLineNumber);
  fmt = va_arg(args, char *);
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  va_end(args);
}


int isValidAdrMode(struct ValidAdrModes * valid, enum AdrMode mode)
{
  int ret = 0;
   
  switch (mode) {
  case am_none:
    ParseError("Interner Fehler: Adressmode NONE.");
    ret = -1;
    break;
  case am_direct:
    ret = valid->am_RegDirect;
    break;
  case am_regind:
    ret = valid->am_RegIndirect;
    break;
  case am_ripostinc:
    ret = valid->am_RegIndirectPostIncr;
    break;
  case am_ripredec:
    ret = valid->am_RegIndirectPreDecr;
    break;
  case am_absolut:
    ret = valid->am_Absolut;
    break;
  case am_pcrel:
    ret = valid->am_PCrel;
    break;
  case am_immed:
    ret = valid->am_Immediate;
    break;
  case am_riadrdist:
    ret = valid->am_RegIndirectAdrDist;
    break;
  default:
    ParseError("Interner Fehler: Unbekannter Adressmode (%x).", mode);
    ret = -1;
  }
  
  return ret;
}

/*
 * -------------------------------------------------------------------------------------
 *
 */


/*
 * Hier kommen die Funktionen, die noch geschrieben/angepasst werden
 * muessen um einen wirklich vollstaendigen Assembler zu bekommen.
 *
 */

int ReadData(void)
{
  int error = 0;
  int xxresult;
  int num;

  /* Diese Funktion tut noch nicht so sehr viel */
  /* es werden nur die Folgeadressen errechnet, */
  /* aber es wird nichts mit den Daten wirklich */
  /* getan. Das muss noch gemacht werden!       */

  do {
    xxresult = yylex();

    switch (xxresult) {
    case xx_STRING:
      CurrentAddress += (strlen(xxStrBuf) + 1) / 2;	/* aufgerundete Laenge in Worten */
      break;
    case xx_NUM:
      CurrentAddress += 1;
      num = strtol(yytext, NULL, 10);
      break;
    case xx_HEX:
      CurrentAddress += 1;
      num = strtol(yytext, NULL, 16);
      break;
    case xx_com:
    case xx_semi:
      break;
    default:
      ParseError("Unerwartetes Token in .DATA Sektion.");
      error = 1;
    }

  } while (xxresult != xx_EOF && xxresult != xx_semi && !error);

  return error;
}


/*
 * Fuer jeden eingelesenen Befehl wird DoCommand() aufgerufen.
 * Als Argumente bekommt die Funktion die Nummer des Kommandos
 * sowie ein Array der Argumente uebergeben.
 *
 * Daraus muss man dann selbst die wirkliche Maschinen-Representation
 * zusammenbauen. Derzeit ist nur als Demo eine Ausgabe der Befehle
 * eingebaut.
 */

int DoCommand(int comNumber, struct ArgAdrMode args[])
{
  int error = 0;
  int i;

  printf("%3d| %-5s Op: %s ", CurrentAddress, ComTab[comNumber].co_name, ComTab[comNumber].co_opcode);
  
  for (i=0; i<ComTab[comNumber].co_NumArgs; i++) {
    switch (args[i].aam_mode) {
    case am_none:
      ParseError("Interner Fehler: Adressmode NONE.");
      break;
    case am_direct:
      printf("%s (direkt Reg %3s)           ", AdrModesCodes[args[i].aam_mode], args[i].aam_ID);
      break;
    case am_regind:
      printf("%s (ind. Reg %3s)             ", AdrModesCodes[args[i].aam_mode], args[i].aam_ID);
      break;
    case am_ripostinc:
      printf("%s (ind post-inc Reg %3s)     ", AdrModesCodes[args[i].aam_mode], args[i].aam_ID);
      break;
    case am_ripredec:
      printf("%s (ind pre-dec Reg %3s)      ", AdrModesCodes[args[i].aam_mode], args[i].aam_ID);
      break;
    case am_absolut:
      printf("%s (absolut %3d)              ", AdrModesCodes[args[i].aam_mode], args[i].aam_hex);
      break;
    case am_pcrel:
      printf("%s (PC relativ %3d)           ", AdrModesCodes[args[i].aam_mode], args[i].aam_hex);
      break;
    case am_immed:
      printf("%s (Konstante %3d)            ", AdrModesCodes[args[i].aam_mode], args[i].aam_hex);
      break;
    case am_riadrdist:
      printf("%s (ind Reg %3s + AdrDist %3d)", AdrModesCodes[args[i].aam_mode], args[i].aam_ID, args[i].aam_hex);
      break;
    default:
      ParseError("Interner Fehler: Unbekannter Adressmode (%x).", args[i].aam_mode);
      error = 1;
    }
    if (i < ComTab[comNumber].co_NumArgs-1) printf(", ");
  }
  printf("\n");

  return error;
}

